#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""Test for ultra supercritical power plant integrated with charge
storage system

"""

__author__ = "Naresh Susarla"

import pytest
from pyomo.environ import (value, Constraint, Var)

from dispatches.case_studies.fossil_case.ultra_supercritical_plant import (
    ultra_supercritical_powerplant as usc)
from dispatches.case_studies.fossil_case.ultra_supercritical_plant.storage \
     import multiperiod_integrated_storage_usc as mp_usc


pmin = int(0.65 * 436) + 1
pmax = 436 + 30


@pytest.fixture(scope="module")
def multiperiod_model():

    # Build multiperiod integrated power plant base model
    m = None
    model = mp_usc.create_usc_model(m, pmin, pmax)
    return model

@pytest.fixture(scope="module")
def initialized_model():

    # Build ultra-supercritical plant base model
    m = None
    model = mp_usc.create_usc_model(m, pmin, pmax)
    mp_usc.usc_custom_init(model)
    return model

@pytest.mark.unit
def test_usc_model(multiperiod_model):

    assert isinstance(multiperiod_model.fs.plant_min_power_eq, Constraint)
    assert isinstance(multiperiod_model.fs.plant_max_power_eq, Constraint)

    assert isinstance(multiperiod_model.fs.previous_power, Var)
    assert value(multiperiod_model.fs.previous_power) == 300

    assert isinstance(multiperiod_model.fs.previous_salt_inventory_hot, Var)
    assert value(multiperiod_model.fs.previous_salt_inventory_hot) == 75000

    assert isinstance(multiperiod_model.fs.salt_inventory_hot, Var)
    assert value(multiperiod_model.fs.salt_inventory_hot) == 75000

    assert isinstance(multiperiod_model.fs.previous_salt_inventory_cold, Var)
    assert value(multiperiod_model.fs.previous_salt_inventory_cold) == 6739292 - 75000

    assert isinstance(multiperiod_model.fs.salt_inventory_cold, Var)
    assert value(multiperiod_model.fs.salt_inventory_cold) == 6739292 - 75000

    assert isinstance(multiperiod_model.fs.constraint_ramp_down, Constraint)
    assert isinstance(multiperiod_model.fs.constraint_ramp_up, Constraint)
    assert isinstance(multiperiod_model.fs.constraint_salt_inventory_hot, Constraint)
    assert isinstance(multiperiod_model.fs.constraint_salt_maxflow_hot, Constraint)
    assert isinstance(multiperiod_model.fs.constraint_salt_maxflow_cold, Constraint)
    assert isinstance(multiperiod_model.fs.constraint_salt_inventory, Constraint)


@pytest.mark.unit
def test_unfix_dof(multiperiod_model):

    # Verify the degrees of freedom are unfixed and the variables are fixed

    mp_usc.usc_unfix_dof(multiperiod_model)

    assert multiperiod_model.fs.hxc.area.fixed
    assert value(multiperiod_model.fs.hxc.area) == 1904
    assert multiperiod_model.fs.hxd.area.fixed
    assert value(multiperiod_model.fs.hxd.area) == 2830
    assert multiperiod_model.fs.hxc.tube_outlet.temperature[0].fixed
    assert value(multiperiod_model.fs.hxc.tube_outlet.temperature[0]) == 831
    assert multiperiod_model.fs.hxd.shell_inlet.temperature[0].fixed
    assert value(multiperiod_model.fs.hxd.shell_inlet.temperature[0]) == 831
    assert multiperiod_model.fs.hxd.shell_outlet.temperature[0].fixed
    assert value(multiperiod_model.fs.hxd.shell_outlet.temperature[0]) == 513.15

@pytest.mark.unit
def test_custom_initialization(initialized_model):

    # Verify the model structure for the model after custom initialization

    assert isinstance(initialized_model.fs.plant_min_power_eq, Constraint)
    assert isinstance(initialized_model.fs.plant_max_power_eq, Constraint)
    assert isinstance(initialized_model.fs.previous_power, Var)
    assert value(initialized_model.fs.previous_power) == 300

    assert isinstance(initialized_model.fs.previous_salt_inventory_hot, Var)
    assert value(initialized_model.fs.previous_salt_inventory_hot) == 75000

    assert isinstance(initialized_model.fs.salt_inventory_hot, Var)
    assert value(initialized_model.fs.salt_inventory_hot) == 75000

    assert isinstance(initialized_model.fs.previous_salt_inventory_cold, Var)
    assert value(initialized_model.fs.previous_salt_inventory_cold) == 6739292 - 75000

    assert isinstance(initialized_model.fs.salt_inventory_cold, Var)
    assert value(initialized_model.fs.salt_inventory_cold) == 6739292 - 75000

    assert isinstance(initialized_model.fs.constraint_ramp_down, Constraint)
    assert isinstance(initialized_model.fs.constraint_ramp_up, Constraint)
    assert isinstance(initialized_model.fs.constraint_salt_inventory_hot, Constraint)
    assert isinstance(initialized_model.fs.constraint_salt_maxflow_hot, Constraint)
    assert isinstance(initialized_model.fs.constraint_salt_maxflow_cold, Constraint)
    assert isinstance(initialized_model.fs.constraint_salt_inventory, Constraint)

@pytest.mark.unit
def test_get_usc_link_variable_pairs():
    assert str('[(blk1.fs.salt_inventory_hot,'
             'blk2.fs.previous_salt_inventory_hot),'
            '(blk1.fs.plant_power_out[0],'
             'blk2.fs.previous_power)]')



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""
This module contains the ConcreteTubeSide model.
"""

# Import Pyomo libraries
from pyomo.environ import Var, Constraint, PositiveReals, units as pyunits

from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume1DBlock,
    UnitModelBlockData,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    useDefault,
)
from idaes.models.unit_models.heat_exchanger import HeatExchangerFlowPattern
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.constants import Constants as c
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog


__author__ = "Konica Mulani, Jaffer Ghouse"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ConcreteTubeSide")
class ConcreteTubeSideData(UnitModelBlockData):
    """ConcreteTubeSide 1D Unit Model Class.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Phase equilibrium term construction flag",
            doc="""Argument to enable phase equilibrium on the shell side.
- True - include phase equilibrium term
- False - do not include phase equilibrium term""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigValue(
            default={},
            description="Arguments for constructing shell property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)""",
        ),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default=useDefault,
            description="Discretization method to use for DAE transformation",
            doc="""Discretization method to use for DAE transformation. See Pyomo
documentation for supported transformations.""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default=useDefault,
            description="Discretization scheme to use for DAE transformation",
            doc="""Discretization scheme to use when transformating domain. See
Pyomo documentation for supported schemes.""",
        ),
    )

    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=20,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=20)""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "flow_type",
        ConfigValue(
            default=HeatExchangerFlowPattern.cocurrent,
            domain=In(HeatExchangerFlowPattern),
            description="Flow configuration of concrete tube",
            doc="""Flow configuration of concrete tube
- HeatExchangerFlowPattern.cocurrent: shell and tube flows from 0 to 1
(default)
- HeatExchangerFlowPattern.countercurrent: shell side flows from 0 to 1
tube side flows from 1 to 0""",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """

        # Call UnitModel.build to setup dynamics
        super().build()

        # dicretisation if not specified.
        if self.config.flow_type == HeatExchangerFlowPattern.cocurrent:

            set_direction_tube = FlowDirection.forward

            if self.config.transformation_method is useDefault:
                _log.warning(
                    "Discretization method was "
                    "not specified for the tube side of the "
                    "co-current concrete tube. "
                    "Defaulting to finite "
                    "difference method on the tube side."
                )
                self.config.transformation_method = "dae.finite_difference"

            if self.config.transformation_scheme is useDefault:
                _log.warning(
                    "Discretization scheme was "
                    "not specified for the tube side of the "
                    "co-current concrete tube. "
                    "Defaulting to backward finite "
                    "difference on the tube side."
                )
                self.config.transformation_scheme = "BACKWARD"
        elif self.config.flow_type == HeatExchangerFlowPattern.countercurrent:
            set_direction_tube = FlowDirection.backward

            if self.config.transformation_method is useDefault:
                _log.warning(
                    "Discretization method was "
                    "not specified for the tube side of the "
                    "counter-current concrete tube. "
                    "Defaulting to finite "
                    "difference method on the tube side."
                )
                self.config.transformation_method = "dae.finite_difference"

            if self.config.transformation_scheme is useDefault:
                _log.warning(
                    "Discretization scheme was "
                    "not specified for the tube side of the "
                    "counter-current concrete tube. "
                    "Defaulting to forward finite "
                    "difference on the tube side."
                )
                self.config.transformation_scheme = "BACKWARD"
        else:
            raise ConfigurationError(
                "{} ConcreteTubeSide only supports cocurrent and "
                "countercurrent flow patterns, but flow_type configuration"
                " argument was set to {}.".format(self.name, self.config.flow_type)
            )

        self.tube = ControlVolume1DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            transformation_method=self.config.transformation_method,
            transformation_scheme=self.config.transformation_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
        )

        self.tube.add_geometry(flow_direction=set_direction_tube)

        self.tube.add_state_blocks(
            information_flow=set_direction_tube,
            has_phase_equilibrium=self.config.has_phase_equilibrium,
        )

        # Populate tube
        self.tube.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=self.config.has_phase_equilibrium,
        )

        self.tube.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_heat_transfer=True,
        )

        self.tube.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        self.tube.apply_transformation()

        # Add Ports for tube side
        self.add_inlet_port(name="tube_inlet", block=self.tube)
        self.add_outlet_port(name="tube_outlet", block=self.tube)

        # Add reference to control volume geometry
        add_object_reference(self, "tube_area", self.tube.area)
        add_object_reference(self, "tube_length", self.tube.length)

        self._make_performance()

    def _make_performance(self):
        """Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        tube_units = self.config.property_package.get_metadata().get_derived_units

        self.d_tube_outer = Var(
            domain=PositiveReals,
            initialize=0.011,
            doc="Outer diameter of tube",
            units=tube_units("length"),
        )
        self.d_tube_inner = Var(
            domain=PositiveReals,
            initialize=0.010,
            doc="Inner diameter of tube",
            units=tube_units("length"),
        )

        self.tube_heat_transfer_coefficient = Var(
            self.flowsheet().config.time,
            self.tube.length_domain,
            domain=PositiveReals,
            initialize=50,
            doc="Heat transfer coefficient",
            units=tube_units("heat_transfer_coefficient"),
        )

        self.temperature_wall = Var(
            self.flowsheet().config.time,
            self.tube.length_domain,
            domain=PositiveReals,
            initialize=298.15,
            units=tube_units("temperature"),
        )

        # Energy transfer between tube wall and tube
        @self.Constraint(
            self.flowsheet().config.time,
            self.tube.length_domain,
            doc="Convective heat transfer",
        )
        def tube_heat_transfer_eq(self, t, x):
            return self.tube.heat[t, x] == self.tube_heat_transfer_coefficient[
                t, x
            ] * c.pi * pyunits.convert(
                self.d_tube_inner, to_units=tube_units("length")
            ) * (
                pyunits.convert(
                    self.temperature_wall[t, x], to_units=tube_units("temperature")
                )
                - self.tube.properties[t, x].temperature
            )

        # Define tube area in terms of tube diameter
        self.area_calc_tube = Constraint(
            expr=4 * self.tube_area
            == c.pi
            * pyunits.convert(self.d_tube_inner, to_units=tube_units("length")) ** 2
        )

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        Initialization routine for the unit (default solver ipopt).

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        solver = get_solver(solver=solver, options=optarg)

        flags_tube = self.tube.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args,
        )

        init_log.info_high("Initialization Step 1 Complete.")

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        self.tube.release_state(flags_tube)

        init_log.info("Initialization Complete.")

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Tube Area"] = self.tube.area
        var_dict["Tube Outer Diameter"] = self.d_tube_outer
        var_dict["Tube Inner Diameter"] = self.d_tube_inner
        var_dict["Tube Length"] = self.tube.length

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {"Tube Inlet": self.tube_inlet, "Tube Outlet": self.tube_outlet,},
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for i, c in self.tube_heat_transfer_eq.items():
            iscale.constraint_scaling_transform(
                c, iscale.get_scaling_factor(self.tube.heat[i], default=1, warning=True)
            )



##############################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
##############################################################################
"""
Project setup with setuptools
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib
import re

# this will come in handy, probably
cwd = pathlib.Path(__file__).parent.resolve()

# Parse long description from README.md file
with open("README.md") as f:
    lines, capture = [], False
    for line in f:
        s = line.strip()
        if re.match(r"#\s*[Aa]bout", s):
            capture = True
        elif re.match("^#", s):
            break
        elif capture is True:
            lines.append(s)
    if lines:
        long_description = " ".join(lines)
    else:
        long_description = "DISPATCHES project"


def read_requirements(input_file):
    """Build list of requirements from a requirements.txt file
    """
    req = []
    for line in input_file:
        s = line.strip()
        c = s.find("#")  # look for comment
        if c != 0:  # no comment (-1) or comment after start (> 0)
            if c > 0:  # strip trailing comment
                s = s[:c]
            req.append(s)
    return req


class SpecialDependencies:
    """
    The following packages require special treatment, as they change rapidly between release cycles.
    Two separate lists of dependencies are kept:
    - for_release: to be used when cutting a release of DISPATCHES
    - for_prerelease: to be used for the prerelease version of DISPATCHES (i.e. the `main` branch, and all PRs targeting it)
    """
    # idaes-pse: for IDAES DMF -dang 12/2020
    for_release = [
        # NOTE: this will fail until this idaes-pse version is available on PyPI
        "idaes-pse==2.0.0b2",
    ]
    for_prerelease = [
        "idaes-pse @ https://github.com/IDAES/idaes-pse/archive/2.0.0b2.zip",
    ]


SPECIAL_DEPENDENCIES = SpecialDependencies.for_release


########################################################################################

setup(
    name="dispatches",
    url="https://github.com/gmlc-dispatches/dispatches",
    version="1.1.0rc0",
    description="GMLC DISPATCHES software tools",
    long_description=long_description,
    long_description_content_type="text/plain",
    author="DISPATCHES team",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="market simulation, chemical engineering, process modeling, hybrid power systems",
    packages=find_packages(),
    python_requires=">=3.7, <4",
    install_requires=[
        "pytest",
        # we use jupyter notebooks
        "jupyter",
        # for visualizing DMF provenance
        "graphviz",
        "gridx-prescient>=2.2.1",
        "nrel-pysam>=3.0.1",
        *SPECIAL_DEPENDENCIES
    ],
    package_data={
        "": ["*.json"],
        "dispatches.tests.data.prescient_5bus": ["*.csv"],
        "dispatches.case_studies.renewables_case.tests": [
            "rts_results_all_prices.npy",
        ],
        "dispatches.case_studies.renewables_case.data": [
           "Wind_Thermal_Dispatch.csv",
           "309_WIND_1-SimulationOutputs.csv",
            "44.21_-101.94_windtoolkit_2012_60min_80m.srw"
        ],
        "dispatches.case_studies.fossil_case.ultra_supercritical_plant": [
            "pfd_ultra_supercritical_pc.svg",
        ],
    },
)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
__author__ = "Radhakrishna Tumbalam Gooty"

"""
Basic tests for the simplified hydrogen tank model
"""

import pytest

# Pyomo imports
from pyomo.environ import (ConcreteModel,
                           TerminationCondition)
from pyomo.util.check_units import assert_units_consistent

# IDAES imports
from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

# DISPATCHES imports
from dispatches.properties.h2_ideal_vap \
    import configuration as h2_ideal_config
from dispatches.unit_models import SimpleHydrogenTank

# Get the default solver for testing
solver = get_solver()


@pytest.fixture()
def build_model():
    # Create the ConcreteModel and the FlowSheetBlock
    m = ConcreteModel(name="H2TankModel")
    m.fs = FlowsheetBlock(dynamic=False)

    # Load thermodynamic package
    m.fs.h2ideal_props = GenericParameterBlock(**h2_ideal_config)

    # Add hydrogen tank
    m.fs.h2_tank = SimpleHydrogenTank(property_package=m.fs.h2ideal_props)

    # Fix the dof of the tank and initialize
    m.fs.h2_tank.inlet.pressure.fix(101325)
    m.fs.h2_tank.inlet.temperature.fix(300)
    m.fs.h2_tank.inlet.flow_mol.fix(25)
    m.fs.h2_tank.inlet.mole_frac_comp[0, "hydrogen"].fix(1)

    m.fs.h2_tank.dt.fix(3600)
    m.fs.h2_tank.tank_holdup_previous.fix(0)
    m.fs.h2_tank.outlet_to_turbine.flow_mol.fix(10)
    m.fs.h2_tank.outlet_to_pipeline.flow_mol.fix(10)
    m.fs.h2_tank.outlet_to_turbine.mole_frac_comp[0, "hydrogen"].fix(1)
    m.fs.h2_tank.outlet_to_pipeline.mole_frac_comp[0, "hydrogen"].fix(1)

    assert degrees_of_freedom(m) == 0

    m.fs.h2_tank.initialize()

    res = solver.solve(m)
    assert res.solver.termination_condition == TerminationCondition.optimal

    return m


@pytest.mark.unit
def test_constraints(build_model):
    m = build_model
    tank = m.fs.h2_tank

    # Constraints
    assert hasattr(tank, "eq_temperature_1")
    assert hasattr(tank, "eq_temperature_2")
    assert hasattr(tank, "eq_pressure_1")
    assert hasattr(tank, "eq_pressure_2")
    assert hasattr(tank, "tank_material_balance")

    # Additional variables
    assert hasattr(tank, "tank_holdup_previous")
    assert hasattr(tank, "tank_holdup")
    assert hasattr(tank, "dt")

    # Inlet and outlet ports
    assert hasattr(tank, "inlet")
    assert hasattr(tank, "outlet_to_turbine")
    assert hasattr(tank, "outlet_to_pipeline")

    assert_units_consistent(m)


@pytest.mark.unit
def test_solution(build_model):
    m = build_model
    tank = m.fs.h2_tank

    assert (tank.outlet_to_pipeline.pressure[0].value ==
            pytest.approx(101325, 1e-2))
    assert (tank.outlet_to_pipeline.temperature[0].value ==
            pytest.approx(300, 1e-2))
    assert (tank.outlet_to_turbine.pressure[0].value ==
            pytest.approx(101325, 1e-2))
    assert (tank.outlet_to_turbine.temperature[0].value ==
            pytest.approx(300, 1e-2))
    assert (tank.tank_holdup[0].value ==
            pytest.approx(3600 * 5, 1e-2))




#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
# Import Pyomo libraries
from pyomo.environ import Var, Param, NonNegativeReals, units as pyunits
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData)
from idaes.core.util.initialization import solve_indexed_blocks
import idaes.logger as idaeslog

import PySAM.Windpower as wind

_log = idaeslog.getLogger(__name__)


# TODO: move to idaes/core/util/config.py
def list_of_list_of_floats(arg):
    '''Domain validator for lists of floats

    Args:
        arg : argument to be cast to list of floats and validated

    Returns:
        List of list of floats
    '''
    try:
        lst = [[float(i) for i in j] for j in arg]
    except TypeError:
        lst = [ListOf(float)(arg), ]
    return lst


def dict_of_list_of_list_of_floats(arg):
    dic = dict()
    for k, v in arg.items():
        dic[k] = list_of_list_of_floats(v)
    return dic


@declare_process_block_class("Wind_Power", doc="Wind plant using turbine powercurve and resource data")
class WindpowerData(UnitModelBlockData):
    """
    Wind plant using turbine powercurve and resource data.
    Unit model to convert wind resource into electricity.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Wind plant does not support dynamic models, thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag",
        doc="""Wind plant does not have defined volume, thus this must be False."""))
    CONFIG.declare("resource_speed", ConfigValue(
        default=None,
        domain=ListOf(float),
        description="List: Wind speed meters per sec",
        doc="For each time in flowsheet's time set, wind speed [m/s]"))

    CONFIG.declare("resource_probability_density", ConfigValue(
        default=None,
        domain=dict_of_list_of_list_of_floats,
        description="Dictionary of Time: List of (wind speed meters per sec, wind degrees clockwise from north, probability)",
        doc="For each time in flowsheet's time set, a probability density function of "
            "Wind speed [m/s] and Wind direction [degrees clockwise from N]"))
    CONFIG.declare("capacity_factor", ConfigValue(
        default=None,
        domain=ListOf(float),
        description="Capacity Factors per Timestep",
        doc="Capacity Factor in a list for each Timestep in model"))

    def build(self):
        """Building model

        Args:
            None
        Returns:
            None
        """
        super().build()

        self.system_capacity = Var(within=NonNegativeReals,
                                     initialize=0.0,
                                     doc="Rated system capacity of wind farm",
                                     units=pyunits.kW)

        self.capacity_factor = Param(self.flowsheet().config.time,
                                     within=NonNegativeReals,
                                     mutable=True,
                                     initialize=0.0,
                                     doc="Ratio of power output to rated capacity, on annual or time series basis",
                                     units=pyunits.kW/pyunits.kW)

        self.electricity = Var(self.flowsheet().config.time,
                               within=NonNegativeReals,
                               initialize=0.0,
                               doc="Electricity production",
                               units=pyunits.kW)

        self.electricity_out = Port(noruleinit=True, doc="A port for electricity flow")
        self.electricity_out.add(self.electricity, "electricity")

        @self.Constraint(self.flowsheet().config.time)
        def elec_from_capacity_factor(b, t):
            return b.electricity[t] <= self.system_capacity * self.capacity_factor[t]

        self.setup_resource()

    def _get_performance_contents(self, time_point=0):
        return {"vars": {"Electricity": self.electricity[time_point].value}}

    @staticmethod
    def setup_atb_turbine():
        """
        """
        wind_simulation = wind.default("WindpowerSingleowner")

        # Use ATB Turbine 2018 Market Average
        wind_simulation.Turbine.wind_turbine_hub_ht = 110
        wind_simulation.Turbine.wind_turbine_rotor_diameter = 116
        wind_simulation.Turbine.wind_turbine_powercurve_powerout = [0, 0, 0, 40.500000, 177.700000, 403.900000, 737.600000, 1187.200000, 1771.100000, 2518.600000, 3448.400000, 4562.500000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 5000, 0, 0]
        wind_simulation.Turbine.wind_turbine_powercurve_windspeeds = [
            1 * i for i in range(len(wind_simulation.Turbine.wind_turbine_powercurve_powerout))]

        # Use a single turbine, do not model wake effects
        wind_simulation.Farm.wind_farm_xCoordinates = [0]
        wind_simulation.Farm.wind_farm_yCoordinates = [0]
        wind_simulation.Farm.system_capacity = max(wind_simulation.Turbine.wind_turbine_powercurve_powerout)
        return wind_simulation

    def setup_resource(self):
        """
        """
        if self.config.resource_probability_density is not None:
            for time in list(self.flowsheet().config.time.data()):
                if time not in self.config.resource_probability_density.keys():
                    raise ValueError("'resource_probability_density' must contain data for time {}".format(time))

            for time, resource in self.config.resource_probability_density.items():
                if abs(sum(r[2] for r in resource) - 1) > 1e-3:
                    raise ValueError("Error in 'resource_probability_density' for time {}: Probabilities of "
                                     "Wind Speed and Direction Probability Density Function must sum to 1")
                if len(resource) != 1:
                    raise NotImplementedError
                wind_simulation = self.setup_atb_turbine()
                wind_simulation.Resource.wind_resource_model_choice = 2
                wind_simulation.Resource.wind_resource_distribution = resource
                wind_simulation.execute(0)
                self.capacity_factor[time].set_value(wind_simulation.Outputs.capacity_factor / 100.)
        elif self.config.resource_speed is not None:
            if len(self.config.resource_speed) < len(self.flowsheet().config.time.data()):
                raise ValueError(f"'resource_speed' list length must be at least {len(self.flowsheet().config.time.data())}")
            for time, speed in zip(list(self.flowsheet().config.time.data()), self.config.resource_speed):
                wind_simulation = self.setup_atb_turbine()
                wind_simulation.Resource.wind_resource_model_choice = 1
                wind_simulation.Resource.weibull_k_factor = 100
                wind_simulation.Resource.weibull_reference_height = 110
                wind_simulation.Resource.weibull_wind_speed = speed
                wind_simulation.execute(0)
                self.capacity_factor[time].set_value(wind_simulation.Outputs.capacity_factor / 100.)
        elif self.config.capacity_factor:
            if not len(self.config.capacity_factor) >= len(self.flowsheet().config.time.data()):
                raise ValueError("Config with 'capacity_factor' must provide data per time step")
                
            for n, time in enumerate(self.flowsheet().config.time.data()):
                self.capacity_factor[time].set_value(self.config.capacity_factor[n])   
        else:
            raise ValueError("Config with 'resource_probability_density' or `capacity_factor` must be provided using `default` argument")

    def initialize_build(self, **kwargs):
        for t in self.flowsheet().config.time:
            self.electricity[t].set_value(self.system_capacity * self.capacity_factor[t])



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
"""
Ideal property package for H2 vapor
"""
# Import Python libraries
import logging

from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import VaporPhase, Component, PhaseType as PT

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal

from idaes.models.properties.modular_properties.pure.NIST import NIST

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal Air hydrogen system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] The NIST Webbook, https://webbook.nist.gov/, retrieved 15th Dec 2020

configuration = {
    # Specifying components
    "components": {
        'hydrogen': {"type": Component,
                     "elemental_composition": {"H": 2},
                     "valid_phase_types": PT.vaporPhase,
                     "cp_mol_ig_comp": NIST,
                     "enth_mol_ig_comp": NIST,
                     "entr_mol_ig_comp": NIST,
                     "parameter_data": {
                         "mw": (2.016e-3, pyunits.kg/pyunits.mol),  # [1]
                         "pressure_crit": (12.9e5, pyunits.Pa),  # [1]
                         "temperature_crit": (33.0, pyunits.K),  # [1]
                         "cp_mol_ig_comp_coeff": {
                             'A': (33.066178,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K),
                             'B': (-11.363417,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK),
                             'C': (11.432816,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK**2),
                             'D': (-2.772874,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK**3),
                             'E': (-0.158558,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K*pyunits.kK**2),
                             'F': (-9.980797,  # [2]
                                   pyunits.J/pyunits.mol),
                             'G': (172.707974,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K),
                             'H': (0,  # [2]
                                   pyunits.kJ/pyunits.mol)}}},
        'nitrogen': {"type": Component,
                     "elemental_composition": {"N": 2},
                     "valid_phase_types": PT.vaporPhase,
                     "cp_mol_ig_comp": NIST,
                     "enth_mol_ig_comp": NIST,
                     "entr_mol_ig_comp": NIST,
                     "parameter_data": {
                         "mw": (28.0134e-3, pyunits.kg / pyunits.mol),  # [1]
                         "pressure_crit": (34.0e5, pyunits.Pa),  # [1]
                         "temperature_crit": (126.15, pyunits.K),  # [1]
                         "cp_mol_ig_comp_coeff": {
                          'A': (19.50583,  # [2]
                                pyunits.J / pyunits.mol / pyunits.K),
                          'B': (19.88705,  # [2]
                                pyunits.J / pyunits.mol / pyunits.K / pyunits.kK),
                          'C': (-8.598535,  # [2]
                                pyunits.J / pyunits.mol / pyunits.K / pyunits.kK ** 2),
                          'D': (1.369784,  # [2]
                                pyunits.J / pyunits.mol / pyunits.K / pyunits.kK ** 3),
                          'E': (0.527601,  # [2]
                                pyunits.J / pyunits.mol / pyunits.K * pyunits.kK ** 2),
                          'F': (-4.935202,  # [2]
                                pyunits.J / pyunits.mol),
                          'G': (212.39000,  # [2]
                                pyunits.J / pyunits.mol / pyunits.K),
                          'H': (0,  # [2]
                                pyunits.kJ / pyunits.mol)}}},
        'oxygen':   {"type": Component,
                     "elemental_composition": {"O": 2},
                     "valid_phase_types": PT.vaporPhase,
                     "cp_mol_ig_comp": NIST,
                     "enth_mol_ig_comp": NIST,
                     "entr_mol_ig_comp": NIST,
                     "parameter_data": {
                         "mw": (31.9988e-3, pyunits.kg/pyunits.mol),  # [1]
                         "pressure_crit": (50.5e5, pyunits.Pa),  # [1]
                         "temperature_crit": (154.55, pyunits.K),  # [1]
                         "cp_mol_ig_comp_coeff": {
                             'A': (31.32234,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K),
                             'B': (-20.23531,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK),
                             'C': (57.86644,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK**2),
                             'D': (-36.50624,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK**3),
                             'E': (-0.007374,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K*pyunits.kK**2),
                             'F': (-8.903471,  # [2]
                                   pyunits.J/pyunits.mol),
                             'G': (246.7945,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K),
                             'H': (0,  # [2]
                                   pyunits.kJ/pyunits.mol)}}},

        'water':  {"type": Component,
                   "elemental_composition": {"H": 2, "O": 1},
                   "valid_phase_types": PT.vaporPhase,
                   "cp_mol_ig_comp": NIST,
                   "enth_mol_ig_comp": NIST,
                   "entr_mol_ig_comp": NIST,
                   "parameter_data": {
                     "mw": (18.0153e-3, pyunits.kg / pyunits.mol),  # [1]
                     "pressure_crit": (220.5e5, pyunits.Pa),  # [1]
                     "temperature_crit": (647.15, pyunits.K),  # [1]
                     "cp_mol_ig_comp_coeff": {
                       'A': (30.092,  # [2]
                             pyunits.J / pyunits.mol / pyunits.K),
                       'B': (6.832514,  # [2]
                             pyunits.J / pyunits.mol / pyunits.K / pyunits.kK),
                       'C': (6.793435,  # [2]
                             pyunits.J / pyunits.mol / pyunits.K / pyunits.kK ** 2),
                       'D': (-2.53448,  # [2]
                             pyunits.J / pyunits.mol / pyunits.K / pyunits.kK ** 3),
                       'E': (0.082139,  # [2]
                             pyunits.J / pyunits.mol / pyunits.K * pyunits.kK ** 2),
                       'F': (-250.881,  # [2]
                             pyunits.J / pyunits.mol),
                       'G': (223.3967,  # [2]
                             pyunits.J / pyunits.mol / pyunits.K),
                       'H': (0.0,  # [2] -241.8264
                             pyunits.kJ / pyunits.mol)}}},
        'argon': {"type": Component,
                  "elemental_composition": {"Ar": 1},
                  "valid_phase_types": PT.vaporPhase,
                  "cp_mol_ig_comp": NIST,
                  "enth_mol_ig_comp": NIST,
                  "entr_mol_ig_comp": NIST,
                  "parameter_data": {
                    "mw": (39.948e-3, pyunits.kg / pyunits.mol),  # [1]
                    "pressure_crit": (48.7e5, pyunits.Pa),  # [1]
                    "temperature_crit": (151.15, pyunits.K),  # [1]
                    "cp_mol_ig_comp_coeff": {
                      'A': (20.786,  # [2]
                            pyunits.J / pyunits.mol / pyunits.K),
                      'B': (0.000000282,  # [2]
                            pyunits.J / pyunits.mol / pyunits.K / pyunits.kK),
                      'C': (-0.000000146,  # [2]
                            pyunits.J / pyunits.mol / pyunits.K / pyunits.kK ** 2),
                      'D': (0.00000001092,  # [2]
                            pyunits.J / pyunits.mol / pyunits.K / pyunits.kK ** 3),
                      'E': (-0.0000000366,  # [2]
                            pyunits.J / pyunits.mol / pyunits.K * pyunits.kK ** 2),
                      'F': (-6.19735,  # [2]
                            pyunits.J / pyunits.mol),
                      'G': (179.999,  # [2]
                            pyunits.J / pyunits.mol / pyunits.K),
                      'H': (0.0,  # [2]
                            pyunits.kJ / pyunits.mol)}}}},
    # Specifying phases
    "phases":  {'Vap': {"type": VaporPhase,
                        "equation_of_state": Ideal}},

    # Set base units of measurement
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},

    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {"flow_mol": (0, 100, 10000, pyunits.mol/pyunits.s),
                     "temperature": (273.15, 300, 2000, pyunits.K),
                     "pressure": (5e4, 1e5, 1e8, pyunits.Pa)},
    "pressure_ref": (101325, pyunits.Pa),  # [2]
    "temperature_ref": (298.15, pyunits.K)}  # [2]
